"""
This type stub file was generated by pyright.
"""

from typing import TYPE_CHECKING
from typing import Any
from typing import Sequence
from typing import TypeVar

from prompt_toolkit import Application
from prompt_toolkit.completion import Completer
from prompt_toolkit.filters import FilterOrBool
from prompt_toolkit.formatted_text import AnyFormattedText
from prompt_toolkit.styles import BaseStyle
from prompt_toolkit.validation import Validator
from prompt_toolkit.widgets import RadioList

if TYPE_CHECKING:
    ...
_T = TypeVar('_T')
class Admonition(RadioList[_T]):
    """Simple scrolling text dialog."""
    open_character = ''
    close_character = ''
    container_style = 'class:admonition-list'
    default_style = 'class:admonition'
    selected_style = 'class:admonition-selected'
    checked_style = 'class:admonition-checked'
    multiple_selection = False
    def __init__(self, values: Sequence[tuple[_T, Any]], default: _T | None = ...) -> None:
        ...
    


def admonition_dialog(title: str = ..., text: str = ..., heading_label: str = ..., ok_text: str | None = ..., cancel_text: str | None = ..., style: BaseStyle | None = ...) -> Application[list[Any]]:
    """Admonition dialog shortcut.
    The focus can be moved between the list and the Ok/Cancel button with tab.
    """
    ...

def input_dialog(title: AnyFormattedText = ..., text: AnyFormattedText = ..., ok_text: str | None = ..., cancel_text: str | None = ..., completer: Completer | None = ..., validator: Validator | None = ..., password: FilterOrBool = ..., style: BaseStyle | None = ..., multiline: bool = ..., default: str = ...) -> Application[str]:
    """
    Display a text input box.
    Return the given text, or None when cancelled.
    """
    ...




"""
This type stub file was generated by pyright.
"""

from typing import TYPE_CHECKING

from prompt_toolkit.validation import Validator

if TYPE_CHECKING:
    ...
class Project:
    def __init__(self, allow_file: list[str] | None = ..., ci_provider: str | None = ..., copyright_head: str | None = ..., enable_cython: bool = ..., enable_uv: bool = ..., github_harden_runner: bool = ..., strict: bool = ..., verify_email: bool = ..., check_package_exists: bool = ...) -> None:
        ...
    
    def __call__(self) -> list[str]:
        ...
    
    def name(self, output: dict[str, list[str]], prefix: dict[str, str]) -> tuple[None | list[str] | str | bool, dict[str, list[str]], dict[str, str]]:
        ...
    
    def summary(self, project_name: str, output: dict[str, list[str]], prefix: dict[str, str]) -> tuple[None | list[str] | str | bool, dict[str, list[str]], dict[str, str]]:
        ...
    
    def keywords(self, project_name: str, output: dict[str, list[str]], prefix: dict[str, str]) -> tuple[None | list[str] | str | bool, dict[str, list[str]], dict[str, str]]:
        ...
    
    def license_file(self, project_name: str, output: dict[str, list[str]], prefix: dict[str, str]) -> tuple[None | list[str] | str | bool, dict[str, list[str]], dict[str, str]]:
        ...
    
    def home_page(self, project_name: str, output: dict[str, list[str]], prefix: dict[str, str]) -> tuple[None | list[str] | str | bool, dict[str, list[str]], dict[str, str]]:
        ...
    
    def author(self, project_name: str, output: dict[str, list[str]], prefix: dict[str, str]) -> tuple[None | list[str] | str | bool, dict[str, list[str]], dict[str, str]]:
        ...
    
    def author_email(self, project_name: str, output: dict[str, list[str]], prefix: dict[str, str]) -> tuple[None | list[str] | str | bool, dict[str, list[str]], dict[str, str]]:
        ...
    
    def license_(self, project_name: str, output: dict[str, list[str]], prefix: dict[str, str]) -> tuple[None | list[str] | str, dict[str, list[str]], dict[str, str]]:
        ...
    
    def license_expression(self, project_name: str, _license: str, output: dict[str, list[str]], prefix: dict[str, str]) -> tuple[None | list[str] | str, dict[str, list[str]], dict[str, str]]:
        ...
    
    def maintainer(self, project_name: str, output: dict[str, list[str]], prefix: dict[str, str]) -> tuple[None | list[str] | str | bool, dict[str, list[str]], dict[str, str]]:
        ...
    
    def maintainer_email(self, project_name: str, output: dict[str, list[str]], prefix: dict[str, str]) -> tuple[None | list[str] | str | bool, dict[str, list[str]], dict[str, str]]:
        ...
    
    def requires_dist(self, project_name: str, output: dict[str, list[str]], prefix: dict[str, str]) -> tuple[list[str] | str | bool | None, dict[str, list[str]], dict[str, str]]:
        ...
    
    def readme_type(self, project_name: str, output: dict[str, list[str]], prefix: dict[str, str]) -> tuple[str | list[str], dict[str, list[str]], dict[str, str]]:
        ...
    
    def typing(self, project_name: str, output: dict[str, list[str]], prefix: dict[str, str]) -> tuple[str | list[str], dict[str, list[str]], dict[str, str]]:
        ...
    
    def project_urls(self, project_name: str, output: dict[str, list[str]], prefix: dict[str, str]) -> tuple[str, dict[str, list[str]], dict[str, str]]:
        ...
    
    def header_input(self, label: str, output: dict[str, list[str]], prefix: dict[str, str], *args: str, validator: Validator | None = ..., split_on: str | None = ...) -> tuple[bool | None | list[str], dict[str, list[str]], dict[str, str],]:
        ...
    





"""
This type stub file was generated by pyright.
"""

ATEXT = ...
ATEXT_RE = ...
DOT_ATOM_TEXT = ...
ATEXT_INTL = ...
ATEXT_INTL_RE = ...
DOT_ATOM_TEXT_INTL = ...
ATEXT_HOSTNAME_INTL = ...
HOSTNAME_LABEL = ...
DOT_ATOM_TEXT_HOSTNAME = ...
DOMAIN_NAME_REGEX = ...
DOMAIN_LITERAL_CHARS = ...
QUOTED_LOCAL_PART_ADDR = ...
QTEXT_INTL = ...
EMAIL_MAX_LENGTH = ...
LOCAL_PART_MAX_LENGTH = ...
DNS_LABEL_LENGTH_LIMIT = ...
DOMAIN_MAX_LENGTH = ...
CASE_INSENSITIVE_MAILBOX_NAMES = ...



from typing import Any
from typing import Dict
from typing import Optional

import dns.exception
import dns.resolver

from .exceptions_types import EmailUndeliverableError


def caching_resolver(*, timeout: Optional[int] = None, cache=None):
    if timeout is None:
        from . import DEFAULT_TIMEOUT

        timeout = DEFAULT_TIMEOUT
    resolver = dns.resolver.Resolver()
    resolver.cache = cache or dns.resolver.LRUCache()  # type: ignore
    resolver.lifetime = timeout  # type: ignore # timeout, in seconds
    return resolver


def validate_email_deliverability(
    domain: str,
    domain_i18n: str,
    timeout: Optional[int] = None,
    dns_resolver=None,
):
    # Check that the domain resolves to an MX record. If there is no MX record,
    # try an A or AAAA record which is a deprecated fallback for deliverability.
    # Raises an EmailUndeliverableError on failure. On success, returns a dict
    # with deliverability information.

    # If no dns.resolver.Resolver was given, get dnspython's default resolver.
    # Override the default resolver's timeout. This may affect other uses of
    # dnspython in this process.
    if dns_resolver is None:
        from . import DEFAULT_TIMEOUT

        if timeout is None:
            timeout = DEFAULT_TIMEOUT
        dns_resolver = dns.resolver.get_default_resolver()
        dns_resolver.lifetime = timeout  # pyright: ignore
    elif timeout is not None:
        raise ValueError("It's not valid to pass both timeout and dns_resolver.")

    deliverability_info: Dict[str, Any] = {}

    try:
        try:
            # Try resolving for MX records (RFC 5321 Section 5).
            response = dns_resolver.resolve(domain, 'MX')

            # For reporting, put them in priority order and remove the trailing dot in the qnames.
            mtas = sorted(
                [
                    (r.preference, str(r.exchange).rstrip('.'))  # pyright: ignore
                    for r in response  # pyright: ignore
                ],
            )

            # RFC 7505: Null MX (0, ".") records signify the domain does not accept email.
            # Remove null MX records from the mtas list (but we've stripped trailing dots,
            # so the 'exchange' is just "") so we can check if there are no non-null MX
            # records remaining.
            mtas = [
                (preference, exchange) for preference, exchange in mtas if exchange != ''
            ]
            if (
                len(mtas) == 0
            ):  # null MX only, if there were no MX records originally a NoAnswer exception would have occurred
                raise EmailUndeliverableError(
                    f'The domain name {domain_i18n} does not accept email.',
                )

            deliverability_info['mx'] = mtas
            deliverability_info['mx_fallback_type'] = None

        except dns.resolver.NoAnswer:
            # If there was no MX record, fall back to an A record. (RFC 5321 Section 5)
            try:
                response = dns_resolver.resolve(domain, 'A')
                deliverability_info['mx'] = [
                    (0, str(r)) for r in response  # pyright: ignore
                ]
                deliverability_info['mx_fallback_type'] = 'A'

            except dns.resolver.NoAnswer:
                # If there was no A record, fall back to an AAAA record.
                # (It's unclear if SMTP servers actually do this.)
                try:
                    response = dns_resolver.resolve(domain, 'AAAA')
                    deliverability_info['mx'] = [
                        (0, str(r)) for r in response  # pyright: ignore
                    ]
                    deliverability_info['mx_fallback_type'] = 'AAAA'

                except dns.resolver.NoAnswer:
                    # If there was no MX, A, or AAAA record, then mail to
                    # this domain is not deliverable, although the domain
                    # name has other records (otherwise NXDOMAIN would
                    # have been raised).
                    raise EmailUndeliverableError(
                        f'The domain name {domain_i18n} does not accept email.',
                    )

            # Check for a SPF (RFC 7208) reject-all record ("v=spf1 -all") which indicates
            # no emails are sent from this domain (similar to a Null MX record
            # but for sending rather than receiving). In combination with the
            # absence of an MX record, this is probably a good sign that the
            # domain is not used for email.
            try:
                response = dns_resolver.resolve(domain, 'TXT')
                for rec in response:  # pyright: ignore
                    value = b''.join(rec.strings)  # pyright: ignore
                    if value.startswith(b'v=spf1 '):
                        deliverability_info['spf'] = value.decode('ascii', errors='replace')
                        if value == b'v=spf1 -all':
                            raise EmailUndeliverableError(
                                f'The domain name {domain_i18n} does not send email.',
                            )
            except dns.resolver.NoAnswer:
                # No TXT records means there is no SPF policy, so we cannot take any action.
                pass

    except dns.resolver.NXDOMAIN:
        # The domain name does not exist --- there are no records of any sort
        # for the domain name.
        raise EmailUndeliverableError(f'The domain name {domain_i18n} does not exist.')

    except dns.resolver.NoNameservers:
        # All nameservers failed to answer the query. This might be a problem
        # with local nameservers, maybe? We'll allow the domain to go through.
        return {
            'unknown-deliverability': 'no_nameservers',
        }

    except dns.exception.Timeout:
        # A timeout could occur for various reasons, so don't treat it as a failure.
        return {
            'unknown-deliverability': 'timeout',
        }

    except EmailUndeliverableError:
        # Don't let these get clobbered by the wider except block below.
        raise

    except Exception as e:
        # Unhandled conditions should not propagate.
        raise EmailUndeliverableError(
            'There was an error while checking if the domain name in the email address is deliverable: '
            + str(e),
        )

    return deliverability_info



from __future__ import annotations

import os
import sys
from io import UnsupportedOperation
from typing import TYPE_CHECKING
from unittest.mock import Mock

from prompt_toolkit.shortcuts.dialogs import button_dialog
from prompt_toolkit.shortcuts.dialogs import checkboxlist_dialog
from prompt_toolkit.shortcuts.dialogs import message_dialog
from prompt_toolkit.shortcuts.dialogs import radiolist_dialog
from prompt_toolkit.shortcuts.dialogs import yes_no_dialog
from tap_producer import TAP

from ozi_core._i18n import TRANSLATION
from ozi_core.fix.build_definition import walk
from ozi_core.fix.missing import get_relpath_expected_files
from ozi_core.ui._style import _style
from ozi_core.ui.menu import MenuButton

if sys.platform != 'win32':  # pragma: no cover
    import curses
else:  # pragma: no cover
    curses = Mock()
    curses.tigetstr = lambda x: b''
    curses.setupterm = lambda: None

if TYPE_CHECKING:  # pragma: no cover
    from argparse import Namespace
    from pathlib import Path


def main_menu(  # pragma: no cover
    output: dict[str, list[str]],
    prefix: dict[str, str],
) -> tuple[None | list[str] | bool, dict[str, list[str]], dict[str, str]]:
    while True:
        match button_dialog(
            title=TRANSLATION('new-dlg-title'),
            text=TRANSLATION('main-menu-text'),
            buttons=[
                MenuButton.RESET._tuple,
                MenuButton.EXIT._tuple,
                MenuButton.BACK._tuple,
            ],
            style=_style,
        ).run():
            case MenuButton.BACK.value:
                break
            case MenuButton.RESET.value:
                if yes_no_dialog(
                    title=TRANSLATION('new-dlg-title'),
                    text=TRANSLATION('main-menu-yn-reset'),
                    style=_style,
                    yes_text=MenuButton.YES._str,
                    no_text=MenuButton.NO._str,
                ).run():
                    return ['interactive', '.'], output, prefix
            case MenuButton.EXIT.value:
                if yes_no_dialog(
                    title=TRANSLATION('new-dlg-title'),
                    text=TRANSLATION('main-menu-yn-exit'),
                    style=_style,
                    yes_text=MenuButton.YES._str,
                    no_text=MenuButton.NO._str,
                ).run():
                    return ['-h'], output, prefix
    return None, output, prefix


class Prompt:
    def __init__(self: Prompt, target: Path) -> None:  # pragma: no cover
        self.target = target
        self.fix: str = ''

    def set_fix_mode(  # pragma: no cover
        self: Prompt,
        project_name: str,
        output: dict[str, list[str]],
        prefix: dict[str, str],
    ) -> tuple[list[str] | str | bool | None, dict[str, list[str]], dict[str, str]]:
        self.fix = radiolist_dialog(
            title=TRANSLATION('fix-dlg-title'),
            text=TRANSLATION('fix-add'),
            style=_style,
            cancel_text=MenuButton.MENU._str,
            values=[('source', 'source'), ('test', 'test'), ('root', 'root')],
        ).run()
        if self.fix is not None:
            output['fix'].append(self.fix)
        return None, output, prefix

    def add_or_remove(  # pragma: no cover
        self: Prompt,
        project_name: str,
        output: dict[str, list[str]],
        prefix: dict[str, str],
    ) -> tuple[list[str] | str | bool | None, dict[str, list[str]], dict[str, str]]:
        add_files: list[str] = []
        output.setdefault('--add', [])
        output.setdefault('--remove', [])
        while True:
            match button_dialog(
                title=TRANSLATION('fix-dlg-title'),
                text='\n'.join(
                    (
                        '\n'.join(iter(prefix)),
                        '\n',
                        TRANSLATION('fix-add-or-remove', projectname=project_name),
                    ),
                ),
                buttons=[
                    MenuButton.ADD._tuple,
                    MenuButton.REMOVE._tuple,
                    MenuButton.MENU._tuple,
                    MenuButton.OK._tuple,
                ],
                style=_style,
            ).run():
                case MenuButton.ADD.value:
                    rel_path, _ = get_relpath_expected_files(self.fix, project_name)
                    files = []
                    with TAP.suppress():
                        for d in walk(self.target, rel_path, []):
                            for k, v in d.items():
                                files += [str(k / i) for i in v['missing']]
                    if len(files) > 0:
                        result = checkboxlist_dialog(
                            title=TRANSLATION('fix-dlg-title'),
                            text='',
                            values=[(i, i) for i in files],
                            style=_style,
                        ).run()
                        if result is not None:
                            add_files += result
                            prefix.update(
                                {
                                    f'Add-{self.fix}: {add_files}': (
                                        f'Add-{self.fix}: {add_files}'
                                    ),
                                },
                            )
                            for f in files:
                                output['--add'].append(f)
                    else:
                        message_dialog(
                            title=TRANSLATION('fix-dlg-title'),
                            text=f'no missing {self.fix} files',
                            style=_style,
                        ).run()
                case MenuButton.REMOVE.value:
                    if len(add_files) != 0:
                        del_files = checkboxlist_dialog(
                            title=TRANSLATION('fix-dlg-title'),
                            text=TRANSLATION('fix-remove'),
                            values=list(zip(add_files, add_files)),
                            style=_style,
                            cancel_text=MenuButton.BACK._str,
                        ).run()
                        if del_files:
                            add_files = list(
                                set(add_files).symmetric_difference(
                                    set(del_files),
                                ),
                            )
                            for f in del_files:
                                output['--remove'].append(f)
                                prefix.update({f'Remove: {f}': f'Remove: {f}'})
                    else:
                        message_dialog(
                            title=TRANSLATION('fix-dlg-title'),
                            text=TRANSLATION('fix-nothing-to-remove'),
                            style=_style,
                            ok_text=MenuButton.OK._str,
                        ).run()
                case MenuButton.OK.value:
                    break
                case MenuButton.MENU.value:
                    result, output, prefix = main_menu(output, prefix)
                    if result is not None:
                        return result, output, prefix
        return None, output, prefix


def interactive_prompt(project: Namespace) -> list[str]:  # pragma: no cover # noqa: C901
    ret_args = ['source']
    try:
        curses.setupterm()
        e3 = curses.tigetstr('E3') or b''
        clear_screen_seq = curses.tigetstr('clear') or b''
        os.write(sys.stdout.fileno(), e3 + clear_screen_seq)
    except UnsupportedOperation:
        pass
    p = Prompt(project.target)
    result, output, prefix = p.set_fix_mode(
        project_name=project.name, output={'fix': []}, prefix={}
    )
    if isinstance(result, list):
        return result
    result, output, prefix = p.add_or_remove(
        project_name=project.name, output=output, prefix=prefix
    )
    if isinstance(result, list):
        return result
    fix = output.pop('fix')
    for k, v in output.items():
        for i in v:
            if len(i) > 0:
                ret_args += [k, i]
    return fix + ret_args + ['.']



"""
This type stub file was generated by pyright.
"""

from __future__ import annotations

import curses
import os
import sys
from argparse import Namespace
from typing import TYPE_CHECKING
from unittest.mock import Mock

from ozi_core.new.interactive.project import Project

"""
``ozi-new`` interactive prompts
"""
if sys.platform != 'win32':
    ...
else:
    ...
if TYPE_CHECKING:
    ...
def interactive_prompt(project: Namespace) -> list[str]:
    ...




# ozi/render.py
# Part of the OZI Project, under the Apache License v2.0 with LLVM Exceptions.
# See LICENSE.txt for license information.
# SPDX-License-Identifier: Apache-2.0 WITH LLVM-exception
"""Rendering utilities for the OZI project templates.
"""
from __future__ import annotations

import configparser
from pathlib import Path
from typing import TYPE_CHECKING
from typing import AnyStr
from typing import Literal
from warnings import warn

from git import InvalidGitRepositoryError
from git import Repo
from ozi_spec import METADATA
from ozi_templates.filter import underscorify  # pyright: ignore
from tap_producer import TAP

from ozi_core._i18n import TRANSLATION

if TYPE_CHECKING:  # pragma: no cover
    from jinja2 import Environment


def find_user_template(target: str, file: str, fix: str) -> str | None:
    """Find a user-defined project template file e.g. :file:`{target}/templates/{fix}/{file}`.

    :param target: path to an OZI project directory
    :type target: Path
    :param file: filename
    :type file: str
    :param fix: template directory fix path
    :type fix: str
    :return: a user-defined template as a string
    :rtype: str | None
    """
    fp = Path(target, 'templates', fix, file)
    if fp.exists():
        with open(fp, encoding='UTF-8') as template:
            user_template = template.read()
    else:
        TAP.ok(TRANSLATION('term-tap-user-template-not-found'), skip=True, template=str(fp))
        user_template = None
    return user_template


def map_to_template(  # noqa: C901
    fix: (
        Literal[
            'child',
            'github_workflows',
            'root',
            'source',
            'subprojects',
            'templates',
            'test',
        ]
        | AnyStr
    ),
    filename: str,
) -> str:
    """Map an appropriate template for an ozi-fix mode and filename.

    .. versionadded:: 1.5

    :param fix: ozi-fix mode setting
    :type fix: AnyStr
    :param filename: name with file extension
    :type filename: str
    :return: template path
    :rtype: str
    """
    match fix, filename:
        case ['test' | 'root', f] if f.endswith('.py'):
            x: str = METADATA.spec.python.src.template.add_test
        case ['source', f] if f.endswith('.py'):
            x = METADATA.spec.python.src.template.add_source
        case ['source', f] if f.endswith('.pyx'):  # pragma: no cover
            x = 'project.name/new_ext.pyx.j2'
        case ['root', f]:
            x = f'{f}.j2'
        case ['source', f]:
            x = f'project.name/{f}.j2'
        case ['test', f]:
            x = f'tests/{f}.j2'
        case ['templates', f]:
            x = f'templates/{f}'
        case ['subprojects', f]:
            x = f'subprojects/{f}.j2'
        case ['child', f]:
            x = 'new_child.j2'
        case ['github_workflows', f]:
            x = f'github_workflows/{f}.j2'
        case [_, _]:  # pragma: no cover
            x = ''
    return x


def build_file(
    env: Environment,
    fix: (
        Literal[
            'child',
            'github_workflows',
            'root',
            'source',
            'subprojects',
            'templates',
            'test',
        ]
        | AnyStr
    ),
    path: Path,
    user_template: str | None,
    **kwargs: str,
) -> None:
    """Render project file based on OZI templates.

    .. versionadded:: 1.5

    :param env: rendering environment
    :type env: Environment
    :param fix: ozi-fix setting
    :type fix: AnyStr
    :param path: full path of file to be rendered
    :type path: Path
    :param user_template: path to a user template to extend
    :type user_template: str | None
    """
    try:
        template = env.get_template(map_to_template(fix, path.name)).render(
            user_template=user_template,
            **kwargs,
        )
        path.parent.mkdir(parents=True, exist_ok=True)
        path.write_text(template)
    except LookupError as e:
        warn(str(e), RuntimeWarning)


def build_child(env: Environment, parent: str, child: Path) -> None:
    """Add a child directory to a parent in an existing OZI-style project.

    :param env: the OZI project file rendering environment
    :type env: jinja2.Environment
    :param parent: existing directory name in project
    :type parent: str
    :param child: path to a new child directory
    :type child: Path
    """
    child.mkdir(parents=True)
    parent = parent.rstrip('/')
    heirs = parent.split('/')
    if len(heirs) > 1:
        warn(
            'Nested folder creation not supported.',
            RuntimeWarning,
            stacklevel=0,
        )
    else:
        build_file(
            env,
            'child',
            (child / 'meson.build'),
            find_user_template(str(parent / child), 'meson.build.j2', 'child'),
            parent=parent,
        )


class RenderedContent:
    def __init__(
        self: RenderedContent,  # pyright: ignore
        env: Environment,
        target: Path,
        name: str,
        ci_provider: str,
        readme_type: str,
    ) -> None:
        """OZI new project content to render.

        .. versionadded:: 1.15.2

        :param target: project root path
        :type target: Path
        :param name: project name
        :type name: str
        :param ci_provider: :term:`CI` provider setting
        :type ci_provider: str
        :param readme_type: the README file extension
        :type readme_type: str
        """
        self.env = env
        self.target = target
        self.name = name
        self.ci_provider = ci_provider
        self.readme_type = readme_type
        self.ci_user = ''

    def render(self: RenderedContent) -> None:
        """Render the project."""
        self.ci_user = render_ci_files_set_user(self.env, self.target, self.ci_provider)
        render_project_files(self.env, self.target, self.name)
        if self.ci_provider == 'github':
            abspath = Path(self.target).resolve()
            Path(
                abspath,
                f'README.{self.readme_type}',
            ).symlink_to(Path('README'))
        else:  # pragma: no cover
            pass


def render_ci_files_set_user(env: Environment, target: Path, ci_provider: str) -> str:
    """Render :term:`CI` files based on the ci_provider for target in env.

    :param env: the OZI project file rendering environment
    :type env: jinja2.Environment
    :param target: directory path to render the project
    :type target: Path
    :param ci_provider: the name of the project continuous integration provider
    :type ci_provider: str
    :return: the ci_user of the target repository for the continuous integration provider
    :rtype: str
    """
    repo = Repo.init(target, mkdir=False)
    try:
        ci_user = repo.config_reader().get('user', 'name')
    except (InvalidGitRepositoryError, configparser.NoSectionError):  # pragma: no cover
        ci_user = ''

    match ci_provider:
        case 'github':
            for filename in METADATA.spec.python.src.template.ci_provider['github']:
                build_file(
                    env,
                    'github_workflows',
                    target / filename.replace('github_workflows', '.github/workflows'),
                    find_user_template(str(target), str(filename), 'github_workflows'),
                )
        case _:  # pragma: no cover
            ci_user = ''
    return ci_user


def render_project_files(env: Environment, target: Path, name: str) -> None:
    """Render the primary new project files(excluding CI).

    :param env: the OZI project file rendering environment
    :type env: jinja2.Environment
    :param target: directory path to render the project
    :type target: Path
    :param name: the canonical project name (without normalization)
    :type name: str
    """
    templates = METADATA.spec.python.src.template
    all_templates = [
        (i, getattr(templates, i))
        for i in ('root', 'test', 'source', 'templates', 'subprojects')
    ]
    for fix, files in all_templates:
        for filename in files:
            filename = (
                filename
                if fix != 'source'
                else filename.replace('project.name', underscorify(name).lower())
            )
            build_file(
                env,
                fix,
                target / filename,
                find_user_template(str(target), filename, fix),
            )



# ozi/actions.py
# Part of the OZI Project, under the Apache License v2.0 with LLVM Exceptions.
# See LICENSE.txt for license information.
# SPDX-License-Identifier: Apache-2.0 WITH LLVM-exception
"""Parsing actions for the OZI commandline interface."""
from __future__ import annotations

import json
import sys
from argparse import Action
from dataclasses import dataclass
from difflib import get_close_matches
from typing import TYPE_CHECKING
from typing import Any
from typing import Collection
from typing import NoReturn
from warnings import warn

import requests
from ozi_spec import METADATA
from ozi_spec import OZI
from ozi_spec import Metadata
from packaging.version import Version
from packaging.version import parse
from pyparsing import ParseException
from spdx_license_list import LICENSES
from tap_producer import TAP

from ozi_core._i18n import TRANSLATION
from ozi_core.spdx import spdx_license_expression
from ozi_core.trove import Prefix
from ozi_core.trove import from_prefix

if TYPE_CHECKING:  # pragma: no cover
    from argparse import ArgumentParser
    from argparse import Namespace
    from collections.abc import Sequence

_prefix = Prefix()


@dataclass
class ExactMatch:
    """Exact matches data for packaging core metadata."""

    audience: tuple[str, ...] = from_prefix(_prefix.audience)
    language: tuple[str, ...] = from_prefix(_prefix.language)
    framework: tuple[str, ...] = from_prefix(_prefix.framework)
    environment: tuple[str, ...] = from_prefix(_prefix.environment)
    license: tuple[str, ...] = from_prefix(_prefix.license)
    license_id: tuple[str, ...] = tuple(
        k for k, v in LICENSES.items() if v.deprecated_id is False
    )
    license_exception_id: tuple[str, ...] = METADATA.spec.python.pkg.license.exceptions
    status: tuple[str, ...] = from_prefix(_prefix.status)
    topic: tuple[str, ...] = from_prefix(_prefix.topic)


class CloseMatch(Action):
    """Special argparse choices action. Warn the user if a close match could not be found."""

    exact_match = ExactMatch()

    def __init__(
        self: CloseMatch,  # pyright: ignore
        option_strings: list[str],
        dest: str,
        nargs: int | str | None = None,
        **kwargs: Any,
    ) -> None:
        """Argparse init"""
        if nargs not in [None, '?']:
            text = 'nargs (other than "?") not allowed'
            raise ValueError(text)

        super().__init__(option_strings, dest, nargs=nargs, **kwargs)

    def close_matches(
        self: CloseMatch,
        key: str,
        value: str,
    ) -> Sequence[str]:
        """Get a close matches for a Python project packaging core metadata key.

        :param key: Python project packaging core metadata key name (normalized)
        :type key: str
        :param value: the value to query a close match for
        :type value: Sequence[str]
        :return: sequence with the best match or an empty sequence
        :rtype: Sequence[str]
        """
        if value is None:
            return []  # pragma: defer to good-first-issue
        no_match = False
        matches: list[str] | str = []
        if hasattr(self.exact_match, key):
            matches = get_close_matches(
                value,
                getattr(self.exact_match, key),
                cutoff=0.40,
            )
            no_match = False if len(matches) else True
        else:  # pragma: no cover
            matches = [value]
            no_match = True
        if no_match:
            warn(
                TRANSLATION('err-no-close-match', key=key, value=value)
                + f'$ ozi-new -l {key}',
                RuntimeWarning,
                stacklevel=0,
            )
        return matches

    def _set_matches(
        self: CloseMatch,
        key: str,
        values: str | Sequence[str],
        namespace: Namespace,
    ) -> None:
        """Set the matches for a key in namespace."""
        match self.nargs:
            case '?':
                setattr(namespace, self.dest, [self.close_matches(key, v) for v in values])
            case _:
                setattr(
                    namespace,
                    self.dest,
                    self.close_matches(
                        key,
                        values if isinstance(values, str) else values[0],
                    ),
                )

    def __call__(
        self: CloseMatch,
        parser: ArgumentParser,
        namespace: Namespace,
        values: str | Sequence[str] | None,
        option_string: str | None = None,
    ) -> None:
        """Find closest matching class attribute."""
        if values is None:  # pragma: no cover
            return
        if option_string is not None:
            key = option_string.lstrip('-').replace('-', '_')
        else:
            key = ''  # pragma: defer to good-first-issue
        self._set_matches(key, values, namespace)


def check_for_update(
    current_version: Version,
    releases: Collection[Version],
) -> None:  # pragma: defer to python
    """Issue a warning if installed version of OZI is not up to date."""
    match max(releases):
        case latest if latest > current_version:
            TAP.not_ok(
                TRANSLATION(
                    'err-new-version',
                    latest=str(latest),
                    currentversion=str(current_version),
                ),
                'https://pypi.org/project/OZI/',
            )
        case latest if latest < current_version:
            TAP.ok(TRANSLATION('term-tap-dev-version'), str(current_version))
        case latest if latest == current_version:
            TAP.ok(TRANSLATION('term-tap-up-to-date'), str(current_version))


def check_version(version: str) -> NoReturn:  # pragma: defer to PyPI
    """Check for a newer version of OZI and exit."""
    response = requests.get('https://pypi.org/pypi/OZI/json', timeout=30)
    match response.status_code:
        case 200:
            check_for_update(
                current_version=parse(version),
                releases=set(map(parse, response.json()['releases'].keys())),
            )
            TAP.end()
        case _:
            TAP.end(
                skip_reason=TRANSLATION(
                    'version-check-failed',
                    status=str(response.status_code),
                ),
            )
    exit(0)


def info(version: str) -> NoReturn:  # pragma: no cover
    """Print all metadata as JSON and exit."""
    sys.exit(print(json.dumps(Metadata(OZI(version)).asdict(), indent=2)))


def list_available(key: str) -> NoReturn:  # pragma: no cover
    """Print a list of valid values for a key and exit."""
    sys.exit(print(*sorted(getattr(ExactMatch, key.replace('-', '_'))), sep='\n'))


def license_expression(expr: str) -> NoReturn:  # pragma: no cover
    """Validate a SPDX license expression."""
    try:
        spdx_license_expression.parse_string(expr, parse_all=True)
        TAP.ok(expr, TRANSLATION('term-parsing-success'))
    except ParseException as e:
        TAP.not_ok(expr, str(e))
    TAP.end()
    exit(0)



"""
This type stub file was generated by pyright.
"""

from typing import Any

from ozi_spec import METADATA

def main_menu(project: Any, output: dict[str, list[str]], prefix: dict[str, str]) -> tuple[None | list[str] | bool, dict[str, list[str]], dict[str, str]]:
    ...

def options_menu(project: Any, output: dict[str, list[str]], prefix: dict[str, str]) -> tuple[None | list[str] | bool, dict[str, list[str]], dict[str, str]]:
    ...

def edit_menu(project: Any, output: dict[str, list[str]], prefix: dict[str, str]) -> tuple[None | list[str] | bool, dict[str, list[str]], dict[str, str]]:
    ...

